<?php declare(strict_types=1);
/**
 * Twig Extension Class: for handling the environment of Twig
 */

namespace Infomaker\Everyware\Twig;

use BadFunctionCallException;
use BadMethodCallException;
use RuntimeException;
use Twig_Lexer;
use Twig_Loader_Chain;
use Twig_LoaderInterface;
use Twig_Parser;
use Twig_SimpleFunction;
use Twig_SourceContextLoaderInterface;

/**
 * Every_Twig
 *
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Twig
 * @since   Infomaker\Everyware\Twig\Every_Twig 0.1
 */
class EveryTwig
{

    /**
     * Contains an instance of the parent for calls to certain functions
     *
     * @var \Twig_Environment
     */
    private static $twig;

    /**
     * Contains an instance of self
     *
     * @var EveryTwig
     */
    private static $instance;

    /**
     * Contains a list of registered loaders
     *
     * @var array
     */
    private static $loaders = [];

    /**
     * Contains whether to use cache
     *
     * @var bool
     */
    protected $cache = false;

    /**
     * Time in seconds to cache the output
     *
     * @var int
     */
    private $cache_time = 60;

    /**
     * Add suffix so we don't have to delete it in deploy every time
     *
     * @var string
     */
    private $cache_key = 'everytwig_compiled_cache_' . GIT_COMMIT;

    private function __construct()
    {
    }

    /**
     * Create new instance if none exist
     *
     * @return EveryTwig
     */
    public static function get_instance(): EveryTwig
    {
        if (static::$instance === null) {
            static::$instance = new static;
        }

        return static::$instance;
    }

    /**
     *Initializes the EveryTwig-wrapper and Twig environment
     *
     * @param array $loaders
     * @param array $functions
     * @param array $filters
     * @param array $options
     *
     * @return Environment
     */
    public function init(
        array $loaders,
        array $functions = [],
        array $filters = [],
        array $options = ['debug' => false, 'cache' => false]
    ): Environment {

        // Handle options
        $this->handle_options($options);

        foreach ($loaders as $loader) {
            $this->add_initial_loader($loader);
        }

        $chain = new Twig_Loader_Chain(static::$loaders);
        $twig = new Environment($chain, $options);

        $twig->addFunction($this->default_php_function());

        foreach ($functions as $function) {
            $twig->addFunction($function);
        }

        foreach ($filters as $filter) {
            $twig->addFilter($filter);
        }

        $twig->setLexer(new Twig_Lexer($twig));
        $twig->setParser(new Twig_Parser($twig));

        return static::$twig = $twig;
    }

    /**
     * Handles the specific cache options and trims them away from the options array.
     *
     * @param array $options
     *
     * @return void
     */
    private function handle_options(array &$options): void
    {

        if (isset($options['cachetime']) && (int)$options['cachetime'] > 0) {
            $this->cache_time = (int)$options['cachetime'];
            unset($options['cachetime']);
        }

        if (isset($options['cache'])) {
            $this->cache = $options['cache'];

            // Add custom cache system to Twig
            $options['cache'] = $this->cache ? ObjectCacheSystem::setup($this->cache_key, $this->cache_time) : false;
        }
    }

    /**
     * Helper function to add a Twig_LoaderInterface loader on init
     *
     * @param Twig_LoaderInterface|Twig_SourceContextLoaderInterface $loader
     *
     * @return void
     */
    private function add_initial_loader(Twig_SourceContextLoaderInterface $loader): void
    {
        static::$loaders[] = $loader;
    }

    /**
     * Twig_SimpleFunction to be able to call default php functions from within twig templates
     *
     * @return Twig_SimpleFunction
     */
    protected function default_php_function(): Twig_SimpleFunction
    {
        $phpFunction = new Twig_SimpleFunction('php_function', function ($name, $args = []) {
            if (!\function_exists($name)) {
                throw new BadFunctionCallException('Function dose not exist');
            }

            if (strpos($name, '::')) {
                $static_method = explode('::', $name);
                if (!method_exists($static_method[0], $static_method[1])) {
                    throw new BadMethodCallException('Method ' . $static_method[1] . ' can not be found on ' . $static_method[0]);
                }
            }

            return \call_user_func_array($name, $args);

        });

        return $phpFunction;
    }

    /**
     * Function to add loaders after a Twig environment has been initialized
     *
     * @param Twig_LoaderInterface|Twig_SourceContextLoaderInterface $loader
     *
     * @throws RuntimeException
     */
    public function add_loader(Twig_SourceContextLoaderInterface $loader): void
    {
        if (static::$twig === null) {
            throw new RuntimeException('Function init needs to be run in order to add extra loaders');
        }

        static::$loaders[] = $loader;
        static::$twig->setLoader(new Twig_Loader_Chain(static::$loaders));
    }

    /**
     * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
     *
     * @param string $template
     * @param array $context
     *
     * @return void
     * @throws \Twig_Error_Loader
     * @throws \Twig_Error_Runtime
     * @throws \Twig_Error_Syntax
     */
    public function render(string $template, array $context = null): void
    {
        static::$twig->display($template, $context ?? []);
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array $context
     *
     * @return mixed
     * @throws \Twig_Error_Loader
     * @throws \Twig_Error_Runtime
     * @throws \Twig_Error_Syntax
     */
    public function view(string $template, array $context = null): string
    {
        return static::$twig->render($template, $context ?? []);
    }
}
