<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Twig_Error;
use Twig_SimpleFilter;
use Twig_SimpleFunction;


abstract class ViewSetup
{
    protected $cache = false;
    protected $cache_time = 86400;
    protected $debug = true;

    /**
     * Contains Wordpress functions that will be added to the list off functions
     * that will be available in the templates
     * function: [func_name] => [custom_name]
     *
     * @var array
     */
    protected $wpFunctions = [];

    /**
     * Contains functions that will be available in the templates.
     * If the same function name is added several times only the last one is kept.
     * @var array
     */
    protected static $functions = [];

    private $filters = [];
    private $globals = [];
    private $loader;

    public function __construct()
    {
        // Wordpress functions
        // ======================================================
        foreach ($this->wpFunctions as $wpFunction => $name) {
            $this->addWpFunction($wpFunction, $name);
        }

        $this->init();
    }


    protected function registerTwigFolder(string $namespace, string $folderPath, array $exclude = null): self
    {
        if (is_dir($folderPath)) {
            try {
                $this->getLoader()->addPathAndSubfolders($folderPath, $namespace, $exclude ?? []);
            } catch (Twig_Error $e) {
                new \Exception($e->getMessage());
            }
        }

        return $this;
    }

    protected function getLoader(): FilesystemLoader
    {
        if ($this->loader === null) {
            $this->loader = new FilesystemLoader();
        }

        return $this->loader;
    }

    protected function addFunction(string $name, $callable, array $options = null): self
    {
        static::$functions[$name] = new Twig_SimpleFunction($name, $callable, $options ?? []);

        return $this;
    }

    protected function addWpFunction(string $functionName, string $name = null)
    {
        $name = (!empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    protected function init(array $options = []): void
    {
        // Use the abstract class in child-classes to customize setup
        $this->setupTwig();

        $twig = EveryTwig::get_instance()->init([$this->getLoader()], static::$functions, $this->filters, array_replace([
            'debug' => $this->debug,
            'cache' => $this->cache,
            'cachetime' => $this->cache_time,
        ], $options));

        //Add global device variable
        if (\defined('EVERY_DEVICE')) {
            $this->addGlobal('device', EVERY_DEVICE);
        }

        foreach ($this->globals as $key => $value) {
            $twig->addGlobal($key, $value);
        }
    }

    abstract public function setupTwig();

    protected function addGlobal(string $key, $value): void
    {
        $this->globals[$key] = $value;
    }

    public static function setup(): ViewSetup
    {
        return new static();
    }

    protected function addFilter(string $name, $callable, array $options = null): self
    {
        $this->filters[] = new Twig_SimpleFilter($name, $callable, $options ?? []);

        return $this;
    }
}
